import time
from typing import Any, Self

from requests.exceptions import RequestException


class MExError(Exception):
    """Base class for generic exceptions."""

    def __str__(self) -> str:
        """Format this exception as a string for logging."""
        args = ", ".join(str(a) for a in self.args) or "N/A"
        return f"{self.__class__.__name__}: {args}"


class EmptySearchResultError(MExError):
    """Empty search result."""


class FoundMoreThanOneError(MExError):
    """Found more than one."""


class MergingError(MExError):
    """Creating a merged item from extracted items and rules failed."""


class TimedRequestException(RequestException):
    """Timed request exception with a seconds attribute."""

    seconds: float

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize exception with timeout `seconds`."""
        self.seconds = kwargs.pop("seconds", 0.0)
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        """Return a shortened representation."""
        args = (
            ", ".join(str(a) for a in self.args)
            or (self.response and self.response.status_code)
            or "N/A"
        )
        return f"{args} (seconds elapsed={self.seconds:.3f})"

    @classmethod
    def create(cls, exc: RequestException, t0: float) -> Self:
        """Create a new timed error from an upstream exception and start time."""
        return cls(
            *exc.args,
            response=exc.response,
            request=exc.request,
            seconds=time.perf_counter() - t0,
        )


class TimedReadTimeout(TimedRequestException):
    """The server did not send any data in the allotted amount of time."""


class TimedTooManyRequests(TimedRequestException):
    """Client sent too many requests in a given time."""


class TimedServerError(TimedRequestException):
    """The server encountered an error or is incapable of responding."""



from mex.common.identity.backend_api import BackendApiIdentityProvider
from mex.common.identity.memory import MemoryIdentityProvider
from mex.common.identity.registry import register_provider
from mex.common.types import IdentityProvider

# register the default providers shipped with mex-common
register_provider(IdentityProvider.MEMORY, MemoryIdentityProvider)
register_provider(IdentityProvider.BACKEND, BackendApiIdentityProvider)



import json
from collections.abc import Generator, Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import TypeGuard

from requests import RequestException

from mex.common.backend_api.connector import BackendApiConnector
from mex.common.logging import logger, watch
from mex.common.models import AnyExtractedModel, AnyMergedModel, AnyRuleSetResponse
from mex.common.settings import BaseSettings
from mex.common.sinks.base import BaseSink
from mex.common.utils import grouper


class BackendApiSink(BaseSink):
    """Sink to load models to the Backend API."""

    CONNECT_TIMEOUT: int | float = 5
    READ_TIMEOUT: int | float = 90

    def __init__(self) -> None:
        """Create a new sink."""
        settings = BaseSettings.get()
        self._executor = ThreadPoolExecutor(
            max_workers=settings.backend_api_parallelization,
            thread_name_prefix="backend_api_sink",
        )

    def close(self) -> None:
        """Close the sink."""
        self._executor.shutdown()

    @watch(log_interval=1000)
    def load(
        self,
        items: Iterable[AnyExtractedModel | AnyMergedModel | AnyRuleSetResponse],
    ) -> Generator[AnyExtractedModel | AnyMergedModel | AnyRuleSetResponse, None, None]:
        """Load extracted models or rule-sets to the Backend API using bulk insertion.

        Args:
            items: Iterable of extracted models or merged models or rule-sets

        Raises:
            NotImplementedError: When you try to load merged items into the backend

        Returns:
            Generator for posted models
        """
        settings = BaseSettings.get()
        futures = [
            self._executor.submit(BackendApiSink.load_chunk, chunk)
            for chunk in grouper(settings.backend_api_chunk_size, items)
        ]
        for future in as_completed(futures):
            yield from future.result()

    @staticmethod
    def is_supported(
        model_list: list[AnyExtractedModel | AnyMergedModel | AnyRuleSetResponse],
    ) -> TypeGuard[list[AnyExtractedModel | AnyRuleSetResponse]]:
        """Return whether the given list only contains models the backend can handle.

        The backend API can only handle extracted models and rule-set responses,
        not merged models. This type guard validates the input and provides
        type narrowing for downstream processing.

        Args:
            model_list: List of models to check for backend compatibility.

        Returns:
            True if all models are supported (extracted models or rule-set responses),
            False otherwise.
        """
        return all(
            isinstance(item, AnyExtractedModel | AnyRuleSetResponse)
            for item in model_list
        )

    @staticmethod
    def load_chunk(
        chunk: Iterable[AnyExtractedModel | AnyMergedModel | AnyRuleSetResponse | None],
    ) -> list[AnyExtractedModel | AnyRuleSetResponse]:
        """Load a chunk of models into the backend.

        Handles the actual bulk insertion of a chunk of models to the backend API.
        Filters out None values and validates that all models are supported before
        attempting ingestion.

        Args:
            chunk: Iterable of models to load, may contain None values.

        Raises:
            NotImplementedError: If the chunk contains unsupported model types.

        Returns:
            List of successfully loaded models.
        """
        connector = BackendApiConnector.get()
        model_list = [item for item in chunk if item is not None]
        if not BackendApiSink.is_supported(model_list):
            msg = "Backend can only ingest extracted models and rule-set responses."
            raise NotImplementedError(msg)
        try:
            connector.ingest(
                model_list,
                timeout=(BackendApiSink.CONNECT_TIMEOUT, BackendApiSink.READ_TIMEOUT),
            )
        except RequestException as error:
            model_info = [
                f"{m.entityType}(hadPrimarySource={m.hadPrimarySource}, "
                f"identifierInPrimarySource={m.identifierInPrimarySource}, ...)"
                if isinstance(m, AnyExtractedModel)
                else f"{m.entityType}(stableTargetId={m.stableTargetId})"
                for m in model_list
            ]
            logger.error(f"Error ingesting models: {', '.join(model_info)}")
            if error.response:
                response_body = json.dumps(error.response.json(), indent=4)
                logger.error(f"Backend responded: {response_body}")
            raise
        return model_list



from typing import Annotated, ClassVar, Literal

from pydantic import Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    ExtractedOrganizationIdentifier,
    MergedOrganizationIdentifier,
    MergedPrimarySourceIdentifier,
    Text,
)

GeprisIdStr = Annotated[
    str,
    Field(
        pattern="^https://gepris\\.dfg\\.de/gepris/institution/[0-9]{1,64}$",
        examples=[
            "https://gepris.dfg.de/gepris/institution/10179",
            "https://gepris.dfg.de/gepris/institution/10293",
            "https://gepris.dfg.de/gepris/institution/21091092",
        ],
        json_schema_extra={"format": "uri"},
    ),
]
GndIdStr = Annotated[
    str,
    Field(
        pattern="^https://d-nb\\.info/gnd/[-X0-9]{3,10}$",
        examples=[
            "https://d-nb.info/gnd/17690-4",
            "https://d-nb.info/gnd/4017909-6",
            "https://d-nb.info/gnd/4603236-8",
        ],
        json_schema_extra={"format": "uri"},
    ),
]
IsniIdStr = Annotated[
    str,
    Field(
        pattern="^https://isni\\.org/isni/[X0-9]{16}$",
        examples=[
            "https://isni.org/isni/0000000109403744",
            "https://isni.org/isni/0000000417918889",
            "https://isni.org/isni/0000000459040795",
        ],
        json_schema_extra={"format": "uri"},
    ),
]
RorIdStr = Annotated[
    str,
    Field(
        pattern="^https://ror\\.org/[a-z0-9]{9}$",
        examples=[
            "https://ror.org/01k5qnb77",
            "https://ror.org/00s9v1h75",
            "https://ror.org/04t3en479",
        ],
        json_schema_extra={"format": "uri"},
    ),
]
ViafIdStr = Annotated[
    str,
    Field(
        pattern="^https://viaf\\.org/viaf/[0-9]{2,22}$",
        examples=[
            "https://viaf.org/viaf/123556639",
            "https://viaf.org/viaf/137080884",
            "https://viaf.org/viaf/122203699",
        ],
        json_schema_extra={"format": "uri"},
    ),
]
WikidataIdStr = Annotated[
    str,
    Field(
        pattern="^http://www\\.wikidata\\.org/entity/[PQ0-9]{2,64}$",
        examples=[
            "http://www.wikidata.org/entity/Q679041",
            "http://www.wikidata.org/entity/Q918501",
            "http://www.wikidata.org/entity/Q491566",
        ],
        json_schema_extra={"format": "uri"},
    ),
]


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["Organization"], Field(frozen=True)]] = (
        "Organization"
    )


class _OptionalLists(_Stem):
    alternativeName: Annotated[
        list[Text],
        Field(
            description="An alternative name for the organization",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/alternative"]},
        ),
    ] = []
    geprisId: Annotated[
        list[GeprisIdStr],
        Field(
            description="Identifier from GEPRIS authority file.",
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P4871"]},
        ),
    ] = []
    gndId: Annotated[
        list[GndIdStr],
        Field(
            description=(
                "An identifier from the German authority file named Gemeinsame "
                "Normdatei (GND), curated by the German National Library (DNB)."
            ),
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P227"]},
        ),
    ] = []
    isniId: Annotated[
        list[IsniIdStr],
        Field(
            description=(
                "The ISNI (International Standard Name Identifier) of the organization."
            ),
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P213"]},
        ),
    ] = []
    rorId: Annotated[
        list[RorIdStr],
        Field(
            description="An identifier of the Research Organization Registry (ROR).",
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P6782"]},
        ),
    ] = []
    shortName: Annotated[
        list[Text],
        Field(
            description="A short name or abbreviation of the organization.",
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P1813"]},
        ),
    ] = []
    viafId: Annotated[
        list[ViafIdStr],
        Field(
            description="Identifier from VIAF (Virtual Authority File).",
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P214"]},
        ),
    ] = []
    wikidataId: Annotated[
        list[WikidataIdStr],
        Field(description="Identifier from Wikidata."),
    ] = []


class _RequiredLists(_Stem):
    officialName: Annotated[
        list[Text],
        Field(
            description="The official name of the organization.",
            min_length=1,
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P1448"]},
        ),
    ]


class _SparseLists(_Stem):
    officialName: Annotated[
        list[Text],
        Field(
            description="The official name of the organization.",
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P1448"]},
        ),
    ] = []


class BaseOrganization(
    _OptionalLists,
    _RequiredLists,
    json_schema_extra={
        "description": (
            "Represents a collection of people organized together into a community or "
            "other social, commercial or political structure. The group has some "
            "common purpose or reason for existence which goes beyond the set of "
            "people belonging to it and can act as an Agent. Organizations are often "
            "decomposable into hierarchical structures ([The Organization Ontology, "
            "2014-01-16](http://www.w3.org/TR/2014/REC-vocab-org-20140116/))."
        ),
        "sameAs": [
            "http://www.w3.org/ns/org#Organization",
            "http://xmlns.com/foaf/0.1/Organization",
            "http://www.w3.org/2006/vcard/ns#Organization",
            "http://www.wikidata.org/entity/Q43229",
        ],
    },
):
    """All fields for a valid organization except for provenance."""


class ExtractedOrganization(BaseOrganization, ExtractedData):
    """An automatically extracted metadata set describing an organization."""

    entityType: Annotated[
        Literal["ExtractedOrganization"], Field(alias="$type", frozen=True)
    ] = "ExtractedOrganization"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(  # noqa: D102
        self,
    ) -> Annotated[
        ExtractedOrganizationIdentifier,
        Field(
            description=(
                "An unambiguous reference to the resource within a given context. "
                "Persistent identifiers should be provided as HTTP URIs "
                "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
            ),
            json_schema_extra={
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"]
            },
        ),
    ]:
        return self._get_identifier(ExtractedOrganizationIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(  # noqa: D102, N802
        self,
    ) -> Annotated[
        MergedOrganizationIdentifier,
        Field(
            description=(
                "The identifier of the merged item that this extracted item belongs to."
            )
        ),
    ]:
        return self._get_stable_target_id(MergedOrganizationIdentifier)


class MergedOrganization(BaseOrganization, MergedItem):
    """The result of merging all extracted items and rules for an organization."""

    entityType: Annotated[
        Literal["MergedOrganization"], Field(alias="$type", frozen=True)
    ] = "MergedOrganization"
    identifier: Annotated[
        MergedOrganizationIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedOrganizationIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class PreviewOrganization(_OptionalLists, _SparseLists, PreviewItem):
    """Preview for merging all extracted items and rules for an organization."""

    entityType: Annotated[
        Literal["PreviewOrganization"], Field(alias="$type", frozen=True)
    ] = "PreviewOrganization"
    identifier: Annotated[
        MergedOrganizationIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedOrganizationIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class AdditiveOrganization(_OptionalLists, _SparseLists, AdditiveRule):
    """Rule to add values to merged organization items."""

    entityType: Annotated[
        Literal["AdditiveOrganization"], Field(alias="$type", frozen=True)
    ] = "AdditiveOrganization"
    supersededBy: Annotated[
        MergedOrganizationIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class SubtractiveOrganization(_OptionalLists, _SparseLists, SubtractiveRule):
    """Rule to subtract values from merged organization items."""

    entityType: Annotated[
        Literal["SubtractiveOrganization"], Field(alias="$type", frozen=True)
    ] = "SubtractiveOrganization"


class PreventiveOrganization(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged organization items."""

    entityType: Annotated[
        Literal["PreventiveOrganization"], Field(alias="$type", frozen=True)
    ] = "PreventiveOrganization"
    alternativeName: list[MergedPrimarySourceIdentifier] = []
    geprisId: list[MergedPrimarySourceIdentifier] = []
    gndId: list[MergedPrimarySourceIdentifier] = []
    isniId: list[MergedPrimarySourceIdentifier] = []
    officialName: list[MergedPrimarySourceIdentifier] = []
    rorId: list[MergedPrimarySourceIdentifier] = []
    shortName: list[MergedPrimarySourceIdentifier] = []
    viafId: list[MergedPrimarySourceIdentifier] = []
    wikidataId: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    """Base class for sets of rules for an organization item."""

    additive: AdditiveOrganization = AdditiveOrganization()
    subtractive: SubtractiveOrganization = SubtractiveOrganization()
    preventive: PreventiveOrganization = PreventiveOrganization()


class OrganizationRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update an organization item."""

    entityType: Annotated[
        Literal["OrganizationRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "OrganizationRuleSetRequest"


class OrganizationRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve an organization item."""

    entityType: Annotated[
        Literal["OrganizationRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "OrganizationRuleSetResponse"
    stableTargetId: MergedOrganizationIdentifier


class OrganizationMapping(_Stem, BaseMapping):
    """Mapping for describing an organization transformation."""

    entityType: Annotated[
        Literal["OrganizationMapping"], Field(alias="$type", frozen=True)
    ] = "OrganizationMapping"
    officialName: Annotated[list[MappingField[list[Text]]], Field(min_length=1)]
    alternativeName: list[MappingField[list[Text]]] = []
    geprisId: list[MappingField[list[GeprisIdStr]]] = []
    gndId: list[MappingField[list[GndIdStr]]] = []
    isniId: list[MappingField[list[IsniIdStr]]] = []
    rorId: list[MappingField[list[RorIdStr]]] = []
    shortName: list[MappingField[list[Text]]] = []
    viafId: list[MappingField[list[ViafIdStr]]] = []
    wikidataId: list[MappingField[list[WikidataIdStr]]] = []


class OrganizationFilter(_Stem, BaseFilter):
    """Class for defining filter rules for organization items."""

    entityType: Annotated[
        Literal["OrganizationFilter"], Field(alias="$type", frozen=True)
    ] = "OrganizationFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



from mex.common.models import ExtractedPerson
from mex.common.orcid.models import OrcidRecord
from mex.common.types import MergedPrimarySourceIdentifier


def transform_orcid_person_to_mex_person(
    orcid_record: OrcidRecord,
    primary_source_id: MergedPrimarySourceIdentifier,
) -> ExtractedPerson:
    """Transforms a single ORCID person to an ExtractedPerson.

    Args:
        orcid_record: OrcidRecord object of a person.
        primary_source_id: Primary source identifier for Orcid.

    Returns:
        ExtractedPerson.
    """
    id_in_primary_source = orcid_record.orcid_identifier.path
    orcid_id = orcid_record.orcid_identifier.uri
    email = [e for emails in orcid_record.person.emails.email for e in emails.email]
    name = orcid_record.person.name
    given_names = None
    family_name = None
    full_name = None
    if name.visibility == "public":
        if name.given_names:
            given_names = name.given_names.value
        if name.family_name:
            family_name = name.family_name.value
    if family_name and given_names:
        full_name = f"{family_name}, {given_names}"
    else:
        full_name = family_name or given_names

    return ExtractedPerson(
        identifierInPrimarySource=id_in_primary_source,
        hadPrimarySource=primary_source_id,
        givenName=given_names,
        familyName=family_name,
        fullName=full_name,
        orcidId=orcid_id,
        email=email,
    )



from typing import Annotated, Generic, TypeVar

from pydantic import BaseModel, Field

from mex.common.types import MergedPrimarySourceIdentifier

_ValueT = TypeVar("_ValueT")
_MappingRuleT = TypeVar("_MappingRuleT")


class MappingRule(BaseModel, Generic[_ValueT], extra="forbid"):
    """Generic mapping rule model."""

    forValues: Annotated[list[str] | None, Field(title="forValues")] = None
    setValues: Annotated[_ValueT | None, Field(title="setValues")] = None
    rule: Annotated[str | None, Field(title="rule")] = None


class MappingField(BaseModel, Generic[_MappingRuleT], extra="forbid"):
    """Generic mapping field model."""

    fieldInPrimarySource: Annotated[
        str | None,
        Field(title="fieldInPrimarySource"),
    ] = None
    locationInPrimarySource: Annotated[
        str | None, Field(title="locationInPrimarySource")
    ] = None
    examplesInPrimarySource: Annotated[
        list[str] | None, Field(title="examplesInPrimarySource")
    ] = None
    mappingRules: Annotated[
        list[MappingRule[_MappingRuleT]], Field(min_length=1, title="mappingRules")
    ]
    comment: Annotated[str | None, Field(title="comment")] = None


class BaseMapping(BaseModel, extra="forbid"):
    """Base class for mapping implementations."""

    hadPrimarySource: Annotated[
        list[MappingField[MergedPrimarySourceIdentifier]],
        Field(min_length=1),
    ]
    identifierInPrimarySource: Annotated[
        list[MappingField[str]],
        Field(min_length=1),
    ]



from typing import Annotated, ClassVar, Literal

from pydantic import Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    ExtractedContactPointIdentifier,
    MergedContactPointIdentifier,
    MergedPrimarySourceIdentifier,
)

EmailStr = Annotated[
    str,
    Field(
        examples=["info@rki.de"],
        pattern="^[^@ \\t\\r\\n]+@[^@ \\t\\r\\n]+\\.[^@ \\t\\r\\n]+$",
        json_schema_extra={"format": "email"},
    ),
]


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["ContactPoint"], Field(frozen=True)]] = (
        "ContactPoint"
    )


class _RequiredLists(_Stem):
    email: Annotated[
        list[EmailStr],
        Field(
            description="The email address associated to the contact point.",
            json_schema_extra={
                "sameAs": [
                    "http://www.w3.org/2006/vcard/ns#hasEmail",
                    "https://schema.org/email",
                ]
            },
            min_length=1,
        ),
    ]


class _SparseLists(_Stem):
    email: Annotated[
        list[EmailStr],
        Field(
            description="The email address associated to the contact point.",
            json_schema_extra={
                "sameAs": [
                    "http://www.w3.org/2006/vcard/ns#hasEmail",
                    "https://schema.org/email",
                ]
            },
        ),
    ] = []


class BaseContactPoint(
    _RequiredLists,
    json_schema_extra={
        "description": "A mail address, where a group of people has access to.",
        "sameAs": ["https://schema.org/ContactPoint"],
    },
):
    """All fields for a valid contact point except for provenance."""


class ExtractedContactPoint(BaseContactPoint, ExtractedData):
    """An automatically extracted metadata set describing a contact point."""

    entityType: Annotated[
        Literal["ExtractedContactPoint"], Field(alias="$type", frozen=True)
    ] = "ExtractedContactPoint"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(  # noqa: D102
        self,
    ) -> Annotated[
        ExtractedContactPointIdentifier,
        Field(
            description=(
                "An unambiguous reference to the resource within a given context. "
                "Persistent identifiers should be provided as HTTP URIs "
                "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
            ),
            json_schema_extra={
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"]
            },
        ),
    ]:
        return self._get_identifier(ExtractedContactPointIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(  # noqa: D102, N802
        self,
    ) -> Annotated[
        MergedContactPointIdentifier,
        Field(
            description=(
                "The identifier of the merged item that this extracted item belongs to."
            )
        ),
    ]:
        return self._get_stable_target_id(MergedContactPointIdentifier)


class MergedContactPoint(BaseContactPoint, MergedItem):
    """The result of merging all extracted items and rules for a contact point."""

    entityType: Annotated[
        Literal["MergedContactPoint"], Field(alias="$type", frozen=True)
    ] = "MergedContactPoint"
    identifier: Annotated[
        MergedContactPointIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedContactPointIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class PreviewContactPoint(_SparseLists, PreviewItem):
    """Preview for merging all extracted items and rules for a contact point."""

    entityType: Annotated[
        Literal["PreviewContactPoint"], Field(alias="$type", frozen=True)
    ] = "PreviewContactPoint"
    identifier: Annotated[
        MergedContactPointIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedContactPointIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class AdditiveContactPoint(_SparseLists, AdditiveRule):
    """Rule to add values to merged contact point items."""

    entityType: Annotated[
        Literal["AdditiveContactPoint"], Field(alias="$type", frozen=True)
    ] = "AdditiveContactPoint"
    supersededBy: Annotated[
        MergedContactPointIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class SubtractiveContactPoint(_SparseLists, SubtractiveRule):
    """Rule to subtract values from merged contact point items."""

    entityType: Annotated[
        Literal["SubtractiveContactPoint"], Field(alias="$type", frozen=True)
    ] = "SubtractiveContactPoint"


class PreventiveContactPoint(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged contact point items."""

    entityType: Annotated[
        Literal["PreventiveContactPoint"], Field(alias="$type", frozen=True)
    ] = "PreventiveContactPoint"
    email: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    """Base class for sets of rules for a contact point item."""

    additive: AdditiveContactPoint = AdditiveContactPoint()
    subtractive: SubtractiveContactPoint = SubtractiveContactPoint()
    preventive: PreventiveContactPoint = PreventiveContactPoint()


class ContactPointRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update a contact point item."""

    entityType: Annotated[
        Literal["ContactPointRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "ContactPointRuleSetRequest"


class ContactPointRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve a contact point item."""

    entityType: Annotated[
        Literal["ContactPointRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "ContactPointRuleSetResponse"
    stableTargetId: MergedContactPointIdentifier


class ContactPointMapping(_Stem, BaseMapping):
    """Mapping for describing a contact point transformation."""

    entityType: Annotated[
        Literal["ContactPointMapping"], Field(alias="$type", frozen=True)
    ] = "ContactPointMapping"
    email: Annotated[list[MappingField[list[EmailStr]]], Field(min_length=1)]


class ContactPointFilter(_Stem, BaseFilter):
    """Class for defining filter rules for contact point items."""

    entityType: Annotated[
        Literal["ContactPointFilter"], Field(alias="$type", frozen=True)
    ] = "ContactPointFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



from pathlib import Path
from typing import Any, Self, cast

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, HttpUrl, SecretStr, model_validator
from pydantic_settings import BaseSettings as PydanticBaseSettings
from pydantic_settings import SettingsConfigDict
from pydantic_settings.sources import ENV_FILE_SENTINEL, DotenvType, EnvSettingsSource

from mex.common.context import SingleSingletonStore
from mex.common.types import AssetsPath, IdentityProvider, Sink, WorkPath

SETTINGS_STORE = SingleSingletonStore["BaseSettings"]()


class BaseSettings(PydanticBaseSettings):
    """Common settings definition class.

    Settings are accessed through a singleton instance of a pydantic settings class.
    The singleton instance can be loaded lazily by calling `BaseSettings.get()`.

    The base settings should only contain options, that are used by common code.
    To add more configuration options for a specific subsystem, create a new subclass
    and define the required fields there. To load a singleton for that subclass,
    simply call `SubsystemSettings.get()`.

    All configuration options should have a speaking name and a clear description.
    The defaults should be set to a value that works with unit tests and must not
    contain any secrets or live URLs that would break unit test isolation.
    """

    model_config = SettingsConfigDict(
        populate_by_name=True,
        env_prefix="mex_",
        env_file=".env",
        env_file_encoding="utf-8",
        env_nested_delimiter="__",
        extra="ignore",
        validate_default=True,
        validate_assignment=True,
    )

    def __init__(
        self,
        _env_file: DotenvType | None = ENV_FILE_SENTINEL,
        _env_file_encoding: str | None = None,
        _env_nested_delimiter: str | None = None,
        _secrets_dir: str | Path | None = None,
        **values: Any,  # noqa: ANN401
    ) -> None:
        """Construct a new settings instance.

        After building from regular settings sources, we look for a configured
        `assets_dir` and then check that folder for a dotenv file as well.
        Because of this dependency from one setting source to another, sadly
        we cannot use `Config.customise_sources`.
        """
        settings_wo_assets_env_file = self._settings_build_values(
            values,
            _env_file=_env_file,
            _env_file_encoding=_env_file_encoding,
            _env_nested_delimiter=_env_nested_delimiter,
            _secrets_dir=_secrets_dir,
        )
        if assets_dir := settings_wo_assets_env_file.get("MEX_ASSETS_DIR"):
            _env_file = Path(assets_dir, ".env")
        super().__init__(
            _env_file=_env_file,
            _env_file_encoding=_env_file_encoding,
            _env_nested_delimiter=_env_nested_delimiter,
            _secrets_dir=_secrets_dir,
            **values,
        )

    @classmethod
    def get(cls) -> Self:
        """Get the current settings instance from singleton store.

        Returns:
            An instance of BaseSettings or a subclass thereof
        """
        return cast("Self", SETTINGS_STORE.load(cls))

    # Note: We need to hardcode the environment variable names for base settings here,
    # otherwise their prefix will get overwritten with those of a specific subclass.

    debug: bool = Field(
        False,  # noqa: FBT003
        alias="pdb",
        description="Jump into post-mortem debugging after any uncaught exception.",
        validation_alias="MEX_DEBUG",
    )
    sink: list[Sink] = Field(
        [Sink.NDJSON],
        description=(
            "Where to send data that is extracted or ingested. Defaults to writing "
            "ndjson files, but can be configured to push to the backend or the graph."
        ),
        validation_alias="MEX_SINK",
    )
    assets_dir: Path = Field(
        Path.cwd() / "assets",
        description=(
            "Path to directory that contains input files treated as read-only, "
            "looks for a folder named `assets` in the current directory by default."
        ),
        validation_alias="MEX_ASSETS_DIR",
    )
    work_dir: Path = Field(
        Path.cwd(),
        description=(
            "Path to directory that stores generated and temporary files. "
            "Defaults to the current working directory."
        ),
        validation_alias="MEX_WORK_DIR",
    )
    identity_provider: IdentityProvider = Field(
        IdentityProvider.MEMORY,
        description="Provider to assign identifiers to new model instances.",
        validation_alias="MEX_IDENTITY_PROVIDER",
    )
    backend_api_url: HttpUrl = Field(
        HttpUrl("http://localhost:8080/"),
        description="MEx backend API url.",
        validation_alias="MEX_BACKEND_API_URL",
    )
    backend_api_key: SecretStr = Field(
        SecretStr("dummy_write_key"),
        description="Backend API key with write access to call POST/PUT endpoints",
        validation_alias="MEX_BACKEND_API_KEY",
    )
    backend_api_parallelization: int = Field(
        1,
        description=(
            "How many simultaneous threads may spin up to load data into the backend."
        ),
        validation_alias="MEX_BACKEND_API_PARALLELIZATION",
    )
    backend_api_chunk_size: int = Field(
        25,
        description="How many items to load into the backend in one chunk.",
        validation_alias="MEX_BACKEND_API_CHUNK_SIZE",
    )
    verify_session: bool | AssetsPath = Field(
        True,  # noqa: FBT003
        description=(
            "Either a boolean that controls whether we verify the server's TLS "
            "certificate, or a path to a CA bundle to use. If a path is given, it can "
            "be either absolute or relative to the `assets_dir`. Defaults to True."
        ),
        validation_alias="MEX_VERIFY_SESSION",
    )
    organigram_path: AssetsPath = Field(
        AssetsPath("raw-data/organigram/organizational_units.json"),
        description=(
            "Path to the JSON file describing the organizational units, "
            "absolute path or relative to `assets_dir`."
        ),
        validation_alias="MEX_ORGANIGRAM_PATH",
    )
    primary_sources_path: AssetsPath = Field(
        AssetsPath("raw-data/primary-sources/primary-sources.json"),
        description=(
            "Path to the JSON file describing the primary sources, "
            "absolute path or relative to `assets_dir`."
        ),
        validation_alias="MEX_PRIMARY_SOURCES_PATH",
    )
    ldap_url: SecretStr = Field(
        SecretStr("ldap://user:pw@ldap:636"),
        description="LDAP server for person queries with authentication credentials. "
        "Must follow format `ldap://user:pw@host:port`, where "
        "`user` is the username, and "
        "`pw` is the password for authenticating against ldap, "
        "`host` is the url of the ldap server, and "
        "`port` is the port of the ldap server.",
        validation_alias="MEX_LDAP_URL",
    )
    ldap_search_base: str = Field(
        "dc=ldapmock,dc=local",
        description="Search base for the ldap connector.",
        validation_alias="MEX_LDAP_SEARCH_BASE",
    )
    wiki_api_url: HttpUrl = Field(
        HttpUrl("http://wikidata/"),
        description="URL of the Wikidata API used to resolve an ID to an organization.",
        validation_alias="MEX_WIKI_API_URL",
    )
    mex_web_user_agent: str = Field(
        "rki/mex",
        description="User agent is sent in request headers to external services.",
        validation_alias="MEX_WEB_USER_AGENT",
    )
    orcid_api_url: HttpUrl = Field(
        HttpUrl("https://orcid"),
        description="URL of orcid api.",
        validation_alias="MEX_ORCID_API_URL",
    )

    def text(self) -> str:
        """Dump the current settings into a readable table.

        Returns:
            Formatted string with all settings key-value pairs.
        """
        dict_ = self.model_dump()
        indent = max(len(key) for key in dict_)
        return "\n".join(
            [
                f"{key.ljust(indent)} "
                f"{', '.join(str(v) for v in val) if isinstance(val, list) else val}"
                for key, val in dict_.items()
            ]
        )

    @classmethod
    def get_env_name(cls, name: str) -> str:
        """Get the name of the environment variable for field with given name.

        Resolves the actual environment variable name that would be used for a
        given field, taking into account case sensitivity and environment prefix
        configuration.

        Args:
            name: The field name to get the environment variable name for.

        Returns:
            The uppercase environment variable name that maps to the field.
        """
        field = cls.model_fields[name]
        env_settings = EnvSettingsSource(
            cls,
            case_sensitive=cls.model_config.get("case_sensitive", False),
            env_prefix=cls.model_config.get("env_prefix", ""),
        )
        env_info = env_settings._extract_field_info(field, name)  # noqa: SLF001
        return env_info[0][1].upper()

    @model_validator(mode="after")
    def resolve_paths(self) -> Self:
        """Resolve AssetPath and WorkPath."""

        def _resolve(model: PydanticBaseModel, _name: str) -> None:
            value = getattr(model, _name)
            if isinstance(value, AssetsPath) and value.is_relative():
                setattr(model, _name, self.assets_dir.resolve() / value)
            elif isinstance(value, WorkPath) and value.is_relative():
                setattr(model, _name, self.work_dir.resolve() / value)
            elif isinstance(value, PydanticBaseModel):
                for sub_model_field_name in type(value).model_fields:
                    _resolve(value, sub_model_field_name)

        for name in type(self).model_fields:
            _resolve(self, name)
        return self

    @model_validator(mode="after")
    def no_settings_as_attributes(self) -> Self:
        """Validate that no attribute inherits from pydantic.BaseSettings."""

        def _validate(model: PydanticBaseModel, _name: str) -> None:
            value = getattr(model, _name)
            if isinstance(value, PydanticBaseSettings):
                msg = (
                    "settings class (inheriting from BaseSettings) has attribute '"
                    f"{_name}' that also inherits from BaseSettings.) Only top level "
                    "may inherit from BaseSettings. Did you mean to inherit from "
                    "BaseModel?"
                )
                raise ValueError(msg)  # noqa: TRY004  stfu ruff, pydantic wants ValueError
            if isinstance(value, PydanticBaseModel):
                for sub_model_field_name in type(value).model_fields:
                    _validate(value, sub_model_field_name)

        for name in type(self).model_fields:
            _validate(self, name)
        return self



from typing import Annotated, TypeVar

from pydantic import Field

from mex.common.models.base.model import BaseModel
from mex.common.transform import camel_to_split
from mex.common.types import (
    ExtractedIdentifier,
    MergedIdentifier,
    MergedPrimarySourceIdentifier,
)

_MergedIdentifierT = TypeVar("_MergedIdentifierT", bound=MergedIdentifier)
_ExtractedIdentifierT = TypeVar("_ExtractedIdentifierT", bound=ExtractedIdentifier)

HadPrimarySource = Annotated[
    MergedPrimarySourceIdentifier,
    Field(
        json_schema_extra={"sameAs": ["http://www.w3.org/ns/prov#hadPrimarySource"]},
        frozen=True,
        description=(
            "A primary source for a topic refers to something produced by some "
            "agent with direct experience and knowledge about the topic, at the "
            "time of the topic's study, without benefit from hindsight. Because "
            "of the directness of primary sources, they 'speak for themselves' "
            "in ways that cannot be captured through the filter of secondary "
            "sources. As such, it is important for secondary sources to reference "
            "those primary sources from which they were derived, so that their "
            "reliability can be investigated. A primary source relation is a "
            "particular case of derivation of secondary materials from their "
            "primary sources. It is recognized that the determination of primary "
            "sources can be up to interpretation, and should be done according to "
            "conventions accepted within the application's domain ([PROV-O, "
            "2013-04-30 ](http://www.w3.org/TR/2013/REC-prov-o-20130430/))."
        ),
    ),
]
IdentifierInPrimarySource = Annotated[
    str,
    Field(
        examples=["123456", "item-501", "D7/x4/zz.final3"],
        min_length=1,
        max_length=1000,
        pattern=r"^[^\n\r]+$",
        frozen=True,
        description="The identifier of the item used in the primary source.",
    ),
]


class ExtractedData(
    BaseModel,
    extra="forbid",
    model_title_generator=lambda m: camel_to_split(m.__name__),
):
    """Base model for all extracted item classes.

    This class adds two important attributes for metadata provenance: `hadPrimarySource`
    and `identifierInPrimarySource`, which are used to uniquely identify an
    item in its original primary source. The attribute `stableTargetId` has to be set
    by each concrete subclass, like `ExtractedPerson`, because it needs to have the
    correct type, e.g. `MergedPersonIdentifier`.

    This class also adds a validator to automatically set identifiers for provenance.
    """

    # The stableTargetId of the primary source, that this item was
    # extracted from. This field is mandatory for all extracted items to
    # aid with data provenance. Extracted primary sources also have this
    # field and are all extracted from a static primary source for MEx.
    # The extracted primary source for MEx has its own merged item as a
    # primary source.
    hadPrimarySource: HadPrimarySource
    # This is the identifier the original item had in its source system.
    # It is only unique amongst items coming from the same system, because
    # identifier formats are likely to overlap between systems.
    # The value for `identifierInPrimarySource` is therefore only unique in
    # composition with `hadPrimarySource`. MEx uses this composite key to
    # assign a stable and globally unique `identifier` per extracted item.
    identifierInPrimarySource: IdentifierInPrimarySource

    def _get_identifier(
        self, identifier_type: type[_ExtractedIdentifierT]
    ) -> _ExtractedIdentifierT:
        """Consult the identity provider to get the `identifier` for this item.

        Args:
            identifier_type: ExtractedIdentifier-subclass to cast the identifier to

        Returns:
            Identifier of the correct type
        """
        # break import cycle, sigh
        from mex.common.identity import get_provider  # noqa: PLC0415

        return identifier_type(
            get_provider()
            .assign(self.hadPrimarySource, self.identifierInPrimarySource)
            .identifier
        )

    def _get_stable_target_id(
        self, identifier_type: type[_MergedIdentifierT]
    ) -> _MergedIdentifierT:
        """Consult the identity provider to get the `stableTargetId` for this item.

        Args:
            identifier_type: MergedIdentifier-subclass to cast the identifier to

        Returns:
            StableTargetId of the correct type
        """
        # break import cycle, sigh
        from mex.common.identity import get_provider  # noqa: PLC0415

        return identifier_type(
            get_provider()
            .assign(self.hadPrimarySource, self.identifierInPrimarySource)
            .stableTargetId
        )



from typing import Annotated, ClassVar, Literal

from pydantic import AfterValidator, Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    AccessRestriction,
    AnonymizationPseudonymization,
    DataProcessingState,
    ExtractedResourceIdentifier,
    Frequency,
    Identifier,
    Language,
    License,
    Link,
    MergedAccessPlatformIdentifier,
    MergedActivityIdentifier,
    MergedBibliographicResourceIdentifier,
    MergedContactPointIdentifier,
    MergedDistributionIdentifier,
    MergedOrganizationalUnitIdentifier,
    MergedOrganizationIdentifier,
    MergedPersonIdentifier,
    MergedPrimarySourceIdentifier,
    MergedResourceIdentifier,
    PersonalData,
    ResourceCreationMethod,
    ResourceTypeGeneral,
    Text,
    Theme,
    Year,
    YearMonth,
    YearMonthDay,
    YearMonthDayTime,
)

ConformsToStr = Annotated[
    str,
    Field(
        examples=[
            "FHIR",
            "LOINC",
            "SNOMED",
            "ICD-10",
        ],
    ),
]
DoiStr = Annotated[
    str,
    Field(
        pattern="^https?://(?:dx\\.)?doi\\.org/[0-9]{2}\\.[0-9]{4,9}[-_.;()/:A-Za-z0-9]{0,256}$",
        examples=[
            "https://doi.org/10.1007/978-1-0716-2441-8_7",
            "https://doi.org/10.2807/1560-7917.ES.2022.27.46.2200849",
            "https://doi.org/10.3389/fmicb.2022.868887",
            "http://dx.doi.org/10.25646/5147",
            "https://doi.org/10.1016/j.vaccine.2022.11.065",
        ],
    ),
]
LoincIdStr = Annotated[
    str,
    Field(
        pattern="^https://loinc\\.org/[-A-Za-z0-9]{2,64}$",
        examples=[
            "https://loinc.org/95209-3",
            "https://loinc.org/LA26211-5",
            "https://loinc.org/96766-1",
        ],
        json_schema_extra={"format": "uri"},
    ),
]
MeshIdStr = Annotated[
    str,
    Field(
        pattern="^http://id\\.nlm\\.nih\\.gov/mesh/[A-Z0-9]{2,64}$",
        examples=[
            "http://id.nlm.nih.gov/mesh/D001604",
            "http://id.nlm.nih.gov/mesh/T025130",
            "http://id.nlm.nih.gov/mesh/D007717",
        ],
        json_schema_extra={"format": "uri"},
    ),
]
TemporalStr = Annotated[
    str,
    Field(
        examples=[
            "2022-01 bis 2022-03",
            "Sommer 2023",
            "nach 2013",
            "1998-2008",
        ],
    ),
]
AnyContactIdentifier = Annotated[
    MergedOrganizationalUnitIdentifier
    | MergedPersonIdentifier
    | MergedContactPointIdentifier,
    AfterValidator(Identifier),
]
MaxTypicalAgeInt = Annotated[
    int,
    Field(
        examples=[
            "99",
            "21",
        ]
    ),
]
MinTypicalAgeInt = Annotated[
    int,
    Field(
        examples=[
            "0",
            "18",
        ]
    ),
]


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["Resource"], Field(frozen=True)]] = "Resource"


class _OptionalLists(_Stem):
    accessPlatform: Annotated[
        list[MergedAccessPlatformIdentifier],
        Field(
            description="A platform from which the resource can be accessed.",
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#accessService"]},
        ),
    ] = []
    alternativeTitle: Annotated[
        list[Text],
        Field(
            description="An alternative name for the resource.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/alternative"]},
        ),
    ] = []
    anonymizationPseudonymization: Annotated[
        list[AnonymizationPseudonymization],
        Field(
            description=(
                "Indicates whether the data has been anonymized and/or pseudonymized."
            )
        ),
    ] = []
    conformsTo: Annotated[
        list[ConformsToStr],
        Field(
            description=(
                "Standards used in the creation, analysis or transmission of the "
                "resource."
            ),
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/conformsTo"]},
        ),
    ] = []
    contributingUnit: Annotated[
        list[MergedOrganizationalUnitIdentifier],
        Field(
            description=(
                "An organizational unit of RKI, that is contributing to the resource."
            ),
            json_schema_extra={
                "subPropertyOf": ["http://purl.org/dc/terms/contributor"]
            },
        ),
    ] = []
    contributor: Annotated[
        list[MergedPersonIdentifier],
        Field(
            description="A person involved in the creation of the resource.",
            json_schema_extra={
                "subPropertyOf": ["http://purl.org/dc/terms/contributor"]
            },
        ),
    ] = []
    creator: Annotated[
        list[MergedPersonIdentifier],
        Field(
            description="A person responsible for making the resource.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/creator"]},
        ),
    ] = []
    description: Annotated[
        list[Text],
        Field(
            description="A text describing the nature of the resource.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/description"]},
        ),
    ] = []
    distribution: Annotated[
        list[MergedDistributionIdentifier],
        Field(description="An available distribution of the resource"),
    ] = []
    documentation: Annotated[
        list[Link],
        Field(
            description="A link to a document documenting the resource.",
            json_schema_extra={
                "subPropertyOf": ["http://purl.org/dc/terms/isReferencedBy"]
            },
        ),
    ] = []
    externalPartner: Annotated[
        list[MergedOrganizationIdentifier],
        Field(
            description=(
                "An external organization that is somehow involved in the "
                "creation of the resource."
            ),
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/contributor"]},
        ),
    ] = []
    hasLegalBasis: Annotated[
        list[Text],
        Field(
            description=(
                "The legal basis used to justify processing of personal data. "
                "Legal basis (plural: legal bases) are defined by legislations "
                "and regulations, whose applicability is usually restricted to "
                "specific jurisdictions which can be represented using "
                "dpv:hasJurisdiction or dpv:hasLaw. Legal basis can be used "
                "without such declarations, e.g. 'Consent', however their "
                "interpretation will require association with a law, e.g. "
                "'EU GDPR'."
            ),
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasLegalBasis"]},
        ),
    ] = []
    hasPurpose: Annotated[
        list[Text],
        Field(
            description=(
                "A free text statement of the purpose of the processing of data "
                "or personal data."
            ),
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasPurpose"]},
        ),
    ] = []
    icd10code: Annotated[
        list[str],
        Field(description="A concept from ICD-10."),
    ] = []
    instrumentToolOrApparatus: Annotated[
        list[Text],
        Field(
            description=(
                "Instrument, tool, or apparatus used in the research, analysis, "
                "observation, or processing of the object that is the subject of "
                "this resource."
            ),
        ),
    ] = []
    isPartOf: Annotated[
        list[MergedResourceIdentifier],
        Field(
            description=(
                "A related resource, in which the described resource is "
                "physically or logically included."
            ),
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/isPartOf"]},
        ),
    ] = []
    keyword: Annotated[
        list[Text],
        Field(
            description="A keyword or tag describing the resource.",
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#keyword"]},
        ),
    ] = []
    language: Annotated[
        list[Language],
        Field(
            description="A language of the resource.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/language"]},
        ),
    ] = []
    loincId: Annotated[
        list[LoincIdStr],
        Field(description="A concept from LOINC."),
    ] = []
    meshId: Annotated[
        list[MeshIdStr],
        Field(description="A concept from MeSH."),
    ] = []
    method: Annotated[
        list[Text],
        Field(
            description=(
                "Method used in the research, analysis, observation or "
                "processing of the object that is subject to the resource."
            ),
        ),
    ] = []
    methodDescription: Annotated[
        list[Text],
        Field(
            description=(
                "The description of the method, that was used to research, "
                "analysis, observation or processing of the object that was "
                "subject to the resource."
            ),
            json_schema_extra={
                "subPropertyOf": ["http://purl.org/dc/terms/description"]
            },
        ),
    ] = []
    populationCoverage: Annotated[
        list[Text],
        Field(
            description=(
                "The type of population common to all subjects of the data collection."
            ),
        ),
    ] = []
    provenance: Annotated[
        list[Text],
        Field(
            description=(
                "A statement about the lineage of a Dataset. Information about "
                "how the data was collected, including methodologies, tools, and "
                "protocols used."
            ),
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/provenance"]},
        ),
    ] = []
    publication: Annotated[
        list[MergedBibliographicResourceIdentifier],
        Field(
            description=(
                "A publication that deals with the research, analysis, "
                "observation or processing of the object that was subject to the "
                "resource, e.g. a research paper."
            ),
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/isReferencedBy"]},
        ),
    ] = []
    publisher: Annotated[
        list[MergedOrganizationIdentifier],
        Field(
            description="The entity responsible for making the item available.",
        ),
    ] = []
    qualityInformation: Annotated[
        list[Text],
        Field(
            description="Some information about the quality of the resource.",
            json_schema_extra={
                "sameAs": ["http://www.w3.org/ns/dqv#hasQualityAnnotation"]
            },
        ),
    ] = []
    resourceCreationMethod: Annotated[
        list[ResourceCreationMethod],
        Field(
            description="Method how the resource was created.",
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = []
    resourceTypeGeneral: Annotated[
        list[ResourceTypeGeneral],
        Field(
            description="General type of the resource.",
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = []
    resourceTypeSpecific: Annotated[
        list[Text],
        Field(
            description=(
                "A term describing the specific nature of the resource. A more "
                "precise term than given by the property 'resourceTypeGeneral'."
            ),
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = []
    rights: Annotated[
        list[Text],
        Field(
            description=(
                "Information about rights held in and over the resource as well "
                "as rights about the possibilities of the usage of the resource."
            ),
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/rights"]},
        ),
    ] = []
    spatial: Annotated[
        list[Text],
        Field(
            description="Spatial coverage of the resource.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/spatial"]},
        ),
    ] = []
    stateOfDataProcessing: Annotated[
        list[DataProcessingState],
        Field(
            description="The processing state of the data, e.g. raw or aggregated.",
        ),
    ] = []


class _RequiredLists(_Stem):
    contact: Annotated[
        list[AnyContactIdentifier],
        Field(
            description="An agent that serves as a contact for the resource.",
            min_length=1,
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#contactPoint"]},
        ),
    ]
    theme: Annotated[
        list[Theme],
        Field(
            description=(
                "A main category of the resource. A resource can have multiple themes."
            ),
            min_length=1,
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#theme"]},
        ),
    ]
    title: Annotated[
        list[Text],
        Field(
            description="The name of the resource.",
            min_length=1,
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/title"]},
        ),
    ]
    unitInCharge: Annotated[
        list[MergedOrganizationalUnitIdentifier],
        Field(
            description=(
                "This property refers to agents who assume responsibility and "
                "accountability for the resource and its appropriate maintenance."
            ),
            min_length=1,
            json_schema_extra={"sameAs": ["http://dcat-ap.de/def/dcatde/maintainer"]},
        ),
    ]


class _SparseLists(_Stem):
    contact: Annotated[
        list[AnyContactIdentifier],
        Field(
            description="An agent that serves as a contact for the resource.",
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#contactPoint"]},
        ),
    ] = []
    theme: Annotated[
        list[Theme],
        Field(
            description=(
                "A main category of the resource. A resource can have multiple themes."
            ),
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#theme"]},
        ),
    ] = []
    title: Annotated[
        list[Text],
        Field(
            description="The name of the resource.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/title"]},
        ),
    ] = []
    unitInCharge: Annotated[
        list[MergedOrganizationalUnitIdentifier],
        Field(
            description=(
                "This property refers to agents who assume responsibility and "
                "accountability for the resource and its appropriate maintenance."
            ),
            json_schema_extra={"sameAs": ["http://dcat-ap.de/def/dcatde/maintainer"]},
        ),
    ] = []


class _OptionalValues(_Stem):
    accrualPeriodicity: Annotated[
        Frequency | None,
        Field(
            description="The frequency with which items are added to a collection.",
            json_schema_extra={
                "sameAs": ["http://purl.org/dc/terms/accrualPeriodicity"]
            },
        ),
    ] = None
    created: Annotated[
        YearMonthDayTime | YearMonthDay | YearMonth | Year | None,
        Field(
            description="Date of creation of the resource",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/created"]},
        ),
    ] = None
    doi: Annotated[
        DoiStr | None,
        Field(description="The Digital Object Identifier (DOI) of the resource."),
    ] = None
    hasPersonalData: Annotated[
        PersonalData | None,
        Field(
            description=(
                "Indicates, if a resource contains data directly or indirectly "
                "associated or related to an individual."
            ),
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasPersonalData"]},
        ),
    ] = None
    license: Annotated[
        License | None,
        Field(
            description=(
                "A legal document giving official permission to do something with "
                "the resource."
            ),
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/license"]},
        ),
    ] = None
    maxTypicalAge: Annotated[
        MaxTypicalAgeInt | None,
        Field(
            description=(
                "Specifies the maximum age of the population within the data "
                "collection, expressed in years."
            )
        ),
    ] = None
    minTypicalAge: Annotated[
        MinTypicalAgeInt | None,
        Field(
            description=(
                "Specifies the minimum age of the population within the data "
                "collection, expressed in years."
            )
        ),
    ] = None
    modified: Annotated[
        YearMonthDayTime | YearMonthDay | YearMonth | Year | None,
        Field(
            description="Date on which the resource was changed.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/modified"]},
        ),
    ] = None
    sizeOfDataBasis: Annotated[
        str | None,
        Field(
            description=(
                "The size of the underlying data basis, e.g. for studies: the "
                "size of the sample."
            )
        ),
    ] = None
    temporal: Annotated[
        YearMonthDayTime | YearMonthDay | YearMonth | Year | TemporalStr | None,
        Field(
            description="Temporal coverage of the resource.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/temporal"]},
        ),
    ] = None
    wasGeneratedBy: Annotated[
        MergedActivityIdentifier | None,
        Field(
            description=(
                "Generation is the completion of production of a new entity by an "
                "activity. This entity did not exist before generation and "
                "becomes available for usage after this generation."
            ),
            json_schema_extra={"sameAs": "http://www.w3.org/ns/prov#wasGeneratedBy"},
        ),
    ] = None


class _RequiredValues(_Stem):
    accessRestriction: Annotated[
        AccessRestriction,
        Field(
            description="Indicates how access to the resource is restricted.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/accessRights"]},
        ),
    ]


class _SparseValues(_Stem):
    accessRestriction: Annotated[
        AccessRestriction | None,
        Field(
            description="Indicates how access to the resource is restricted.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/accessRights"]},
        ),
    ] = None


class _VariadicValues(_Stem):
    accessRestriction: Annotated[
        list[AccessRestriction],
        Field(
            description="Indicates how access to the resource is restricted.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/accessRights"]},
        ),
    ] = []
    accrualPeriodicity: Annotated[
        list[Frequency],
        Field(
            description="The frequency with which items are added to a collection.",
            json_schema_extra={
                "sameAs": ["http://purl.org/dc/terms/accrualPeriodicity"]
            },
        ),
    ] = []
    created: Annotated[
        list[YearMonthDayTime | YearMonthDay | YearMonth | Year],
        Field(
            description="Date of creation of the resource",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/created"]},
        ),
    ] = []
    doi: Annotated[
        list[DoiStr],
        Field(description="The Digital Object Identifier (DOI) of the resource."),
    ] = []
    hasPersonalData: Annotated[
        list[PersonalData],
        Field(
            description=(
                "Indicates, if a resource contains data directly or indirectly "
                "associated or related to an individual."
            ),
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasPersonalData"]},
        ),
    ] = []
    license: Annotated[
        list[License],
        Field(
            description=(
                "A legal document giving official permission to do something with "
                "the resource."
            ),
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/license"]},
        ),
    ] = []
    maxTypicalAge: Annotated[
        list[MaxTypicalAgeInt],
        Field(
            description=(
                "Specifies the maximum age of the population within the data "
                "collection, expressed in years."
            )
        ),
    ] = []
    minTypicalAge: Annotated[
        list[MinTypicalAgeInt],
        Field(
            description=(
                "Specifies the minimum age of the population within the data "
                "collection, expressed in years."
            )
        ),
    ] = []
    modified: Annotated[
        list[YearMonthDayTime | YearMonthDay | YearMonth | Year],
        Field(
            description="Date on which the resource was changed.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/modified"]},
        ),
    ] = []
    sizeOfDataBasis: Annotated[
        list[str],
        Field(
            description=(
                "The size of the underlying data basis, e.g. for studies: the "
                "size of the sample."
            )
        ),
    ] = []
    temporal: Annotated[
        list[YearMonthDayTime | YearMonthDay | YearMonth | Year | TemporalStr],
        Field(
            description="Temporal coverage of the resource.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/temporal"]},
        ),
    ] = []
    wasGeneratedBy: Annotated[
        list[MergedActivityIdentifier],
        Field(
            description=(
                "Generation is the completion of production of a new entity by an "
                "activity. This entity did not exist before generation and "
                "becomes available for usage after this generation."
            ),
            json_schema_extra={"sameAs": "http://www.w3.org/ns/prov#wasGeneratedBy"},
        ),
    ] = []


class BaseResource(
    _OptionalLists,
    _RequiredLists,
    _OptionalValues,
    _RequiredValues,
    json_schema_extra={
        "description": (
            "A defined piece of information or collection of information on Public "
            "Health, that has been generated as part of a (research) activity at the "
            "RKI or to comply with a (federal) law or regulation that applies to the "
            "RKI."
        ),
        "sameAs": ["http://www.w3.org/ns/dcat#Dataset"],
    },
):
    """All fields for a valid resource except for provenance."""


class ExtractedResource(BaseResource, ExtractedData):
    """An automatically extracted metadata set describing a resource."""

    entityType: Annotated[
        Literal["ExtractedResource"], Field(alias="$type", frozen=True)
    ] = "ExtractedResource"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(  # noqa: D102
        self,
    ) -> Annotated[
        ExtractedResourceIdentifier,
        Field(
            description=(
                "An unambiguous reference to the resource within a given context. "
                "Persistent identifiers should be provided as HTTP URIs "
                "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
            ),
            json_schema_extra={
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"]
            },
        ),
    ]:
        return self._get_identifier(ExtractedResourceIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(  # noqa: D102, N802
        self,
    ) -> Annotated[
        MergedResourceIdentifier,
        Field(
            description=(
                "The identifier of the merged item that this extracted item belongs to."
            )
        ),
    ]:
        return self._get_stable_target_id(MergedResourceIdentifier)


class MergedResource(BaseResource, MergedItem):
    """The result of merging all extracted items and rules for a resource."""

    entityType: Annotated[
        Literal["MergedResource"], Field(alias="$type", frozen=True)
    ] = "MergedResource"
    identifier: Annotated[
        MergedResourceIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedResourceIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class PreviewResource(
    _OptionalLists, _SparseLists, _OptionalValues, _SparseValues, PreviewItem
):
    """Preview for merging all extracted items and rules for a resource."""

    entityType: Annotated[
        Literal["PreviewResource"], Field(alias="$type", frozen=True)
    ] = "PreviewResource"
    identifier: Annotated[
        MergedResourceIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedResourceIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class AdditiveResource(
    _OptionalLists, _SparseLists, _OptionalValues, _SparseValues, AdditiveRule
):
    """Rule to add values to merged resource items."""

    entityType: Annotated[
        Literal["AdditiveResource"], Field(alias="$type", frozen=True)
    ] = "AdditiveResource"
    supersededBy: Annotated[
        MergedResourceIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class SubtractiveResource(
    _OptionalLists, _SparseLists, _VariadicValues, SubtractiveRule
):
    """Rule to subtract values from merged resource items."""

    entityType: Annotated[
        Literal["SubtractiveResource"], Field(alias="$type", frozen=True)
    ] = "SubtractiveResource"


class PreventiveResource(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged resource items."""

    entityType: Annotated[
        Literal["PreventiveResource"], Field(alias="$type", frozen=True)
    ] = "PreventiveResource"
    accessPlatform: list[MergedPrimarySourceIdentifier] = []
    accessRestriction: list[MergedPrimarySourceIdentifier] = []
    accrualPeriodicity: list[MergedPrimarySourceIdentifier] = []
    alternativeTitle: list[MergedPrimarySourceIdentifier] = []
    anonymizationPseudonymization: list[MergedPrimarySourceIdentifier] = []
    conformsTo: list[MergedPrimarySourceIdentifier] = []
    contact: list[MergedPrimarySourceIdentifier] = []
    contributingUnit: list[MergedPrimarySourceIdentifier] = []
    contributor: list[MergedPrimarySourceIdentifier] = []
    created: list[MergedPrimarySourceIdentifier] = []
    doi: list[MergedPrimarySourceIdentifier] = []
    creator: list[MergedPrimarySourceIdentifier] = []
    description: list[MergedPrimarySourceIdentifier] = []
    distribution: list[MergedPrimarySourceIdentifier] = []
    documentation: list[MergedPrimarySourceIdentifier] = []
    externalPartner: list[MergedPrimarySourceIdentifier] = []
    hasLegalBasis: list[MergedPrimarySourceIdentifier] = []
    hasPurpose: list[MergedPrimarySourceIdentifier] = []
    hasPersonalData: list[MergedPrimarySourceIdentifier] = []
    icd10code: list[MergedPrimarySourceIdentifier] = []
    instrumentToolOrApparatus: list[MergedPrimarySourceIdentifier] = []
    isPartOf: list[MergedPrimarySourceIdentifier] = []
    keyword: list[MergedPrimarySourceIdentifier] = []
    language: list[MergedPrimarySourceIdentifier] = []
    license: list[MergedPrimarySourceIdentifier] = []
    loincId: list[MergedPrimarySourceIdentifier] = []
    maxTypicalAge: list[MergedPrimarySourceIdentifier] = []
    meshId: list[MergedPrimarySourceIdentifier] = []
    method: list[MergedPrimarySourceIdentifier] = []
    methodDescription: list[MergedPrimarySourceIdentifier] = []
    minTypicalAge: list[MergedPrimarySourceIdentifier] = []
    modified: list[MergedPrimarySourceIdentifier] = []
    populationCoverage: list[MergedPrimarySourceIdentifier] = []
    provenance: list[MergedPrimarySourceIdentifier] = []
    publication: list[MergedPrimarySourceIdentifier] = []
    publisher: list[MergedPrimarySourceIdentifier] = []
    qualityInformation: list[MergedPrimarySourceIdentifier] = []
    resourceCreationMethod: list[MergedPrimarySourceIdentifier] = []
    resourceTypeGeneral: list[MergedPrimarySourceIdentifier] = []
    resourceTypeSpecific: list[MergedPrimarySourceIdentifier] = []
    rights: list[MergedPrimarySourceIdentifier] = []
    sizeOfDataBasis: list[MergedPrimarySourceIdentifier] = []
    spatial: list[MergedPrimarySourceIdentifier] = []
    stateOfDataProcessing: list[MergedPrimarySourceIdentifier] = []
    temporal: list[MergedPrimarySourceIdentifier] = []
    theme: list[MergedPrimarySourceIdentifier] = []
    title: list[MergedPrimarySourceIdentifier] = []
    unitInCharge: list[MergedPrimarySourceIdentifier] = []
    wasGeneratedBy: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    """Base class for sets of rules for a resource item."""

    additive: AdditiveResource = AdditiveResource()
    subtractive: SubtractiveResource = SubtractiveResource()
    preventive: PreventiveResource = PreventiveResource()


class ResourceRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update a resource item."""

    entityType: Annotated[
        Literal["ResourceRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "ResourceRuleSetRequest"


class ResourceRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve a resource item."""

    entityType: Annotated[
        Literal["ResourceRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "ResourceRuleSetResponse"
    stableTargetId: MergedResourceIdentifier


class ResourceMapping(_Stem, BaseMapping):
    """Mapping for describing a resource transformation."""

    entityType: Annotated[
        Literal["ResourceMapping"], Field(alias="$type", frozen=True)
    ] = "ResourceMapping"
    accessRestriction: Annotated[
        list[MappingField[AccessRestriction]], Field(min_length=1)
    ]
    accrualPeriodicity: list[MappingField[Frequency | None]] = []
    created: list[
        MappingField[YearMonthDayTime | YearMonthDay | YearMonth | Year | None]
    ] = []
    doi: list[MappingField[DoiStr | None]] = []
    hasPersonalData: list[MappingField[PersonalData | None]] = []
    license: list[MappingField[License | None]] = []
    maxTypicalAge: list[MappingField[MaxTypicalAgeInt | None]] = []
    minTypicalAge: list[MappingField[MinTypicalAgeInt | None]] = []
    modified: list[
        MappingField[YearMonthDayTime | YearMonthDay | YearMonth | Year | None]
    ] = []
    sizeOfDataBasis: list[MappingField[str | None]] = []
    temporal: list[
        MappingField[
            YearMonthDayTime | YearMonthDay | YearMonth | Year | TemporalStr | None
        ]
    ] = []
    wasGeneratedBy: list[MappingField[MergedActivityIdentifier | None]] = []
    contact: Annotated[
        list[MappingField[list[AnyContactIdentifier]]],
        Field(min_length=1),
    ]
    theme: Annotated[list[MappingField[list[Theme]]], Field(min_length=1)]
    title: Annotated[list[MappingField[list[Text]]], Field(min_length=1)]
    unitInCharge: Annotated[
        list[MappingField[list[MergedOrganizationalUnitIdentifier]]],
        Field(min_length=1),
    ]
    accessPlatform: list[MappingField[list[MergedAccessPlatformIdentifier]]] = []
    alternativeTitle: list[MappingField[list[Text]]] = []
    anonymizationPseudonymization: list[
        MappingField[list[AnonymizationPseudonymization]]
    ] = []
    conformsTo: list[MappingField[list[ConformsToStr]]] = []
    contributingUnit: list[MappingField[list[MergedOrganizationalUnitIdentifier]]] = []
    contributor: list[MappingField[list[MergedPersonIdentifier]]] = []
    creator: list[MappingField[list[MergedPersonIdentifier]]] = []
    description: list[MappingField[list[Text]]] = []
    distribution: list[MappingField[list[MergedDistributionIdentifier]]] = []
    documentation: list[MappingField[list[Link]]] = []
    externalPartner: list[MappingField[list[MergedOrganizationIdentifier]]] = []
    hasLegalBasis: list[MappingField[list[Text]]] = []
    hasPurpose: list[MappingField[list[Text]]] = []
    icd10code: list[MappingField[list[str]]] = []
    instrumentToolOrApparatus: list[MappingField[list[Text]]] = []
    isPartOf: list[MappingField[list[MergedResourceIdentifier]]] = []
    keyword: list[MappingField[list[Text]]] = []
    language: list[MappingField[list[Language]]] = []
    loincId: list[MappingField[list[LoincIdStr]]] = []
    meshId: list[MappingField[list[MeshIdStr]]] = []
    method: list[MappingField[list[Text]]] = []
    methodDescription: list[MappingField[list[Text]]] = []
    populationCoverage: list[MappingField[list[Text]]] = []
    provenance: list[MappingField[list[Text]]] = []
    publication: list[MappingField[list[MergedBibliographicResourceIdentifier]]] = []
    publisher: list[MappingField[list[MergedOrganizationIdentifier]]] = []
    qualityInformation: list[MappingField[list[Text]]] = []
    resourceCreationMethod: list[MappingField[list[ResourceCreationMethod]]] = []
    resourceTypeGeneral: list[MappingField[list[ResourceTypeGeneral]]] = []
    resourceTypeSpecific: list[MappingField[list[Text]]] = []
    rights: list[MappingField[list[Text]]] = []
    spatial: list[MappingField[list[Text]]] = []
    stateOfDataProcessing: list[MappingField[list[DataProcessingState]]] = []


class ResourceFilter(_Stem, BaseFilter):
    """Class for defining filter rules for resource items."""

    entityType: Annotated[
        Literal["ResourceFilter"], Field(alias="$type", frozen=True)
    ] = "ResourceFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



from typing import cast

from mex.common.connector.http import HTTPConnector
from mex.common.exceptions import EmptySearchResultError
from mex.common.settings import BaseSettings


class WikidataAPIConnector(HTTPConnector):
    """Connector class to handle requesting the Wikidata API."""

    def _set_url(self) -> None:
        """Set url of the host."""
        settings = BaseSettings.get()
        self.url = str(settings.wiki_api_url)

    def _set_session(self) -> None:
        """Create and set request session."""
        super()._set_session()
        settings = BaseSettings.get()
        self.session.headers.update(
            {
                "User-Agent": settings.mex_web_user_agent,
                "Api-User-Agent": settings.mex_web_user_agent,
            }
        )

    def _check_availability(self) -> None:
        """Send a GET request to verify the host is available."""
        self.request("GET", params={"format": "json", "action": "wbgetentities"})

    def get_wikidata_item_details_by_id(self, item_id: str) -> dict[str, str]:
        """Get details of a wikidata item by item id.

        Args:
            item_id: wikidata item id

        Returns:
            Details of the found item.
        """
        results = self.request(
            "GET",
            params={
                "action": "wbgetentities",
                "format": "json",
                "ids": item_id,
                "props": (
                    "info|aliases|labels|descriptions|datatype|claims|"
                    "sitelinks|sitelinks/urls"
                ),
                "formatversion": "2",
            },
        )
        if entities := results.get("entities"):
            return cast("dict[str, str]", entities[item_id])
        if (error := results.get("error")) and error["code"] == "no-such-entity":
            raise EmptySearchResultError(error["info"])
        msg = f"unexpected error getting wikidata item: {item_id}"
        raise RuntimeError(msg)
